#
#  Random Utilities
#

package Util;
use strict;
require Exporter;
our @ISA    = qw(Exporter);
our @EXPORT = qw(diag_error diag_warning diag_info diag_perror
                 util_trim util_chomp util_clone util_var util_split util_simplify_path
                 util_globs_to_regexp util_add_uniquely util_format_variable util_bool
                 cmd_set cmd_init cmd_val cmd_bool cmd_handle_args);

# Diagnostics
sub diag_error(@);
sub diag_warning(@);
sub diag_info(@);
sub diag_perror(@);

# Utilities
sub util_trim($);
sub util_chomp($);
sub util_clone($);
sub util_var($);
sub util_simplify_path($);
sub util_globs_to_regexp(@);
sub util_add_uniquely(\@\@);
sub util_format_variable( $$$;$$ );
sub util_quote(@);

# Command Line handling
sub cmd_set($$);
sub cmd_init();
sub cmd_val($;$);
sub cmd_bool($;$);
sub cmd_handle_args(@);

############################### Diagnostics ###############################

# Print error and exit
sub diag_error(@) {
    print STDERR "ERROR: ", @_, "\n";
    exit 1;
}

# Print I/O error (like C 'perror') and exit
sub diag_perror(@) {
    print STDERR @_, ": $!\n";
    exit 1;
}

# Print warning
sub diag_warning(@) {
    print STDERR "WARNING: ", @_, "\n"
      unless cmd_bool('quiet');
}

# Print information message
sub diag_info(@) {
    print @_, "\n"
      if cmd_bool('verbose');
}

################################ Utilities ################################

# Remove leading/trailing whitespace
sub util_trim($) {
    my $x = shift;
    $x =~ s/^\s+//;
    $x =~ s/\s+$//;
    $x;
}

# Generalized version of chomp() to remove all sorts of linefeeds
sub util_chomp($) {
    my $x = shift;
    $x =~ s/[\r\n]+$//;
    $x;
}

# Clone an object (deep copy). Clones arrays, hashes, and everything
# that has a clone() method.
sub util_clone($) {
    my $x = shift;
    if (ref($x) eq 'ARRAY') {
        my $result = [];
        foreach (@$x) { push @$result, util_clone($_) }
        $result;
    } elsif (ref($x) eq 'HASH') {
        my $result = {};
        foreach (keys %$x) { $result->{$_} = util_clone($x->{$_}) }
        $result;
    } elsif (ref($x)) {
        $x->clone();
    } else {
        $x;
    }
}

# Given a variable name, return Make syntax reference to that variable
sub util_var($) {
    my $x = shift;
    if ((length($x) == 1 && $x !~ /[a-z]/i) || $x eq '**') {
        "\$$x"
    } else {
        "\$($x)"
    }
}

# Given a piece of text, split it into semi-words. That is, the text will
# be split into space-separated words, but $(...) will be kept even
# if they contain spaces.
sub util_split($) {
    my $s = shift;
    my @result = ('');

    pos($s) = 0;
    while (1) {
        if ($s =~ m/\G\s+/gc) {
            push @result, ''
              if $result[-1] ne '';
        } elsif ($s =~ m/\G([^\$\s]+)/gc) {
            $result[-1] .= $1
        } elsif ($s =~ m/\G(\$\()/gc) {
            # $(...)
            my $name = $1;
            my $brace = 1;
            while ($brace) {
                if ($s =~ m/\G\(/gc) {
                    $name .= "("; ++$brace
                } elsif ($s =~ m/\G\)/gc) {
                    $name .= ")"; --$brace;
                } elsif ($s =~ m/\G([^()]+)/gc) {
                    $name .= $1;
                } else {
                    diag_error("malformed variable reference: '$s' unbalanced parens");
                }
            }
            $result[-1] .= $name;
        } elsif ($s =~ m/\G(\$.)/gc) {
            $result[-1] .= $1;
        } else {
            last
        }
    }
    pop @result if $result[-1] eq '';
    @result;
}

# Simplify a path name. This strips "./" sequences as well as
# "/foo/../" groups, which have no effect upon interpretation of the
# path name.
sub util_simplify_path($) {
    my $s = shift;
    while ($s =~ s|^\./||) { }
    while ($s =~ s|/\./|/|) { }
    while ($s =~ s|/[^/]*/\.\./|/|) { }
    $s;
}

# Convert list of globs into regexp matching all the globs
sub util_globs_to_regexp(@) {
    my @out;
    foreach my $g (@_) {
        my $re;
        my $glob = $g;
        pos($glob) = 0;
        while (1) {
            if ($glob =~ m|\G$|sgc) {
                last;
            } elsif ($glob =~ m|\G([\w\s\d]+)|sgc) {
                $re .= $1;
            } elsif ($glob =~ m|\G\*+|sgc) {
                $re .= ".*";
            } elsif ($glob =~ m|\G\?|sgc) {
                $re .= ".";
            } elsif ($glob =~ m|\G(\\.)|sgc) {
                $re .= $1;
            } elsif ($glob =~ m|\G(.)|sgc) {
                $re .= "\\$1";
            } else {
                die "huh? '" . substr($glob, pos($glob)) . "'";
            }
        }
        push @out, $re;
    }
    "(?:" . join('|', @out) . ")";
}

# Quote things for the shell
sub util_quote(@) {
    my @out;
    foreach (@_) {
        my $p = $_;
        $p =~ s|([^-A-Z0-9_.])|\\$1|g;
        push @out, $p;
    }
    @out;
}

# Add elements of one list to another one, but don't produce dupes
sub util_add_uniquely(\@\@) {
    my $acc = shift;
    my $vals = shift;
    foreach my $me (@$vals) {
        push @$acc, $me
          unless grep {$me eq $_} @$acc
    }
}

# Format a variable assignment sensibly
sub util_format_variable( $$$;$$ ) {
    my ($name, $op, $value, $minlength, $wraplength) = @_;
    $minlength ||= 80;
    $wraplength ||= 75;
    my $accum = "$name $op";
    my $result = "";
    if (length($value) > $minlength) {
        foreach (util_split($value)) {
            if (length($accum) + length($_) > $wraplength) {
                $result .= "$accum \\\n";
                $accum = "   ";
            }
            $accum .= " $_";
        }
        $result . $accum . "\n";
    } else {
        $accum . " $value\n";
    }
}

# Interpret user input as boolean
sub util_bool( $ ) {
    my $value = shift;
    return defined($value) && ($value eq '1' || $value =~ /^y/i || $value =~ /^t/i);
}

########################## Command Line Handling ##########################

my %cmd_args;

# Set a command line parameter.
sub cmd_set($$) {
    my ($name, $value) = @_;
    $cmd_args{$name} = $value;
}

# Initialize, after command line has been processed.
sub cmd_init() {
    $cmd_args{'quiet'} ||= "no";
    $cmd_args{'verbose'} ||= "no";
    $cmd_args{'root'}  ||= ".";
    $cmd_args{'dir'}   ||= "$cmd_args{'root'}/P9";

    $cmd_args{root} = util_simplify_path($cmd_args{root});
    $cmd_args{dir} = util_simplify_path($cmd_args{dir});

    diag_info("init: base directory is '$cmd_args{'root'}'");
    diag_info("init: project directory is '$cmd_args{'dir'}'");
}

# Get value of a command line argument. Optional second parameter
# specifies default.
sub cmd_val($;$) {
    my ($name, $default) = @_;
    exists ($cmd_args{$name}) ? $cmd_args{$name} : $default;
}

# Like cmd_val, but conveniently returns a bool
sub cmd_bool($;$) {
    my ($name, $default) = @_;
    my $value = cmd_val($name, $default);
    return util_bool($value);
}

# Process arguments intermixed in a list. I.e., given
#    foo dodge=this bar
# this will set dodge=this, and return a list containing foo, bar.
sub cmd_handle_args(@) {
    my @rv;
    foreach (@_) {
        if (/^(root|dir)=/) {
            diag_error ("cannot set `$1' here, must be before command name");
        } elsif (/^(.*?)=(.*)/) {
            $cmd_args{$1} = $2;
        } else {
            push @rv, $_;
        }
    }
    @rv;
}

1;
